@@ -1,7 +1,7 @@
 //! Domain specific language for the Lazy api.
 use crate::logical_plan::Context;
 use crate::prelude::*;
-use crate::utils::{has_expr, has_wildcard};
+use crate::utils::{expr_to_root_column_name, has_expr, has_wildcard};
 use polars_core::prelude::*;
 
 #[cfg(feature = "temporal")]
@@ -248,6 +248,7 @@ pub enum Expr {
     Alias(Box<Expr>, Arc<String>),
     Column(Arc<String>),
     Columns(Vec<String>),
+    DtypeColumn(Vec<DataType>),
     Literal(LiteralValue),
     BinaryExpr {
         left: Box<Expr>,
@@ -442,6 +443,7 @@ impl fmt::Debug for Expr {
             KeepName(e) => write!(f, "KEEP NAME {:?}", e),
             SufPreFix { expr, .. } => write!(f, "SUF-PREFIX {:?}", expr),
             Columns(names) => write!(f, "COLUMNS({:?})", names),
+            DtypeColumn(dt) => write!(f, "COLUMN OF DTYPE: {:?}", dt),
         }
     }
 }
@@ -458,9 +460,12 @@ pub enum Operator {
     Minus,
     Multiply,
     Divide,
+    #[cfg(feature = "true_div")]
+    TrueDivide,
     Modulus,
     And,
     Or,
+    Xor,
 }
 
 pub fn binary_expr(l: Expr, op: Operator, r: Expr) -> Expr {
@@ -769,17 +774,19 @@ impl Expr {
 
     /// Get unique values of this expression.
     pub fn unique(self) -> Self {
-        if has_expr(&self, |e| matches!(e, Expr::Wildcard)) {
-            panic!("wildcard not supperted in unique expr");
-        }
+        assert!(
+            !has_expr(&self, |e| matches!(e, Expr::Wildcard)),
+            "wildcard not supperted in unique expr"
+        );
         self.apply(|s: Series| s.unique(), GetOutput::same_type())
     }
 
     /// Get the first index of unique values of this expression.
     pub fn arg_unique(self) -> Self {
-        if has_expr(&self, |e| matches!(e, Expr::Wildcard)) {
-            panic!("wildcard not supported in unique expr");
-        }
+        assert!(
+            !has_expr(&self, |e| matches!(e, Expr::Wildcard)),
+            "wildcard not supported in unique expr"
+        );
         self.apply(
             |s: Series| s.arg_unique().map(|ca| ca.into_series()),
             GetOutput::from_type(DataType::UInt32),
@@ -788,9 +795,10 @@ impl Expr {
 
     /// Get the index values that would sort this expression.
     pub fn arg_sort(self, reverse: bool) -> Self {
-        if has_expr(&self, |e| matches!(e, Expr::Wildcard)) {
-            panic!("wildcard not supported in unique expr");
-        }
+        assert!(
+            !has_expr(&self, |e| matches!(e, Expr::Wildcard)),
+            "wildcard not supported in unique expr"
+        );
         self.apply(
             move |s: Series| Ok(s.argsort(reverse).into_series()),
             GetOutput::from_type(DataType::UInt32),
@@ -998,6 +1006,15 @@ impl Expr {
         )
     }
 
+    /// Get an array with the cumulative product computed at every element
+    #[cfg_attr(docsrs, doc(cfg(feature = "cum_agg")))]
+    pub fn cumprod(self, reverse: bool) -> Self {
+        self.apply(
+            move |s: Series| Ok(s.cumprod(reverse)),
+            GetOutput::same_type(),
+        )
+    }
+
     /// Get an array with the cumulative min computed at every element
     #[cfg_attr(docsrs, doc(cfg(feature = "cum_agg")))]
     pub fn cummin(self, reverse: bool) -> Self {
@@ -1039,6 +1056,13 @@ impl Expr {
         self.map(move |s: Series| s.round(decimals), GetOutput::same_type())
     }
 
+    /// Floor underlying floating point array to the lowest integers smaller or equal to the float value.
+    #[cfg(feature = "round_series")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "round_series")))]
+    pub fn floor(self) -> Self {
+        self.map(move |s: Series| s.floor(), GetOutput::same_type())
+    }
+
     /// Apply window function over a subgroup.
     /// This is similar to a groupby + aggregation + self join.
     /// Or similar to [window functions in Postgres](https://www.postgresql.org/docs/9.1/tutorial-window.html).
@@ -1060,7 +1084,7 @@ impl Expr {
     ///      .lazy()
     ///      .select(&[
     ///          col("groups"),
-    ///          sum("values").over(vec![col("groups")]),
+    ///          sum("values").over([col("groups")]),
     ///      ])
     ///      .collect()?;
     ///     dbg!(&out);
@@ -1098,16 +1122,16 @@ impl Expr {
     /// │ 1      ┆ 16     │
     /// ╰────────┴────────╯
     /// ```
-    pub fn over(self, partition_by: Vec<Expr>) -> Self {
+    pub fn over<E: AsRef<[Expr]>>(self, partition_by: E) -> Self {
         Expr::Window {
             function: Box::new(self),
-            partition_by,
+            partition_by: partition_by.as_ref().to_vec(),
             order_by: None,
             options: WindowOptions { explode: false },
         }
     }
 
-    /// Shift the values in the array by some period. See [the eager implementation](polars_core::series::SeriesTrait::fill_null).
+    /// Replace the null values by a value.
     pub fn fill_null(self, fill_value: Expr) -> Self {
         map_binary_lazy_field(
             self,
@@ -1117,8 +1141,8 @@ impl Expr {
                     Ok(a)
                 } else {
                     let st = get_supertype(a.dtype(), b.dtype())?;
-                    let a = a.cast_with_dtype(&st)?;
-                    let b = b.cast_with_dtype(&st)?;
+                    let a = a.cast(&st)?;
+                    let b = b.cast(&st)?;
                     let mask = a.is_not_null();
                     a.zip_with_same_type(&mask, &b)
                 }
@@ -1129,6 +1153,11 @@ impl Expr {
             },
         )
     }
+
+    /// Replace the floating point `NaN` values by a value.
+    pub fn fill_nan(self, fill_value: Expr) -> Self {
+        when(self.clone().is_nan()).then(fill_value).otherwise(self)
+    }
     /// Count the values of the Series
     /// or
     /// Get counts of the group by operation.
@@ -1163,6 +1192,11 @@ impl Expr {
         binary_expr(self, Operator::And, expr)
     }
 
+    // xor operation
+    pub fn xor(self, expr: Expr) -> Self {
+        binary_expr(self, Operator::Xor, expr)
+    }
+
     /// or operation
     pub fn or(self, expr: Expr) -> Self {
         binary_expr(self, Operator::Or, expr)
@@ -1194,35 +1228,37 @@ impl Expr {
     #[cfg(feature = "is_in")]
     #[cfg_attr(docsrs, doc(cfg(feature = "is_in")))]
     pub fn is_in(self, other: Expr) -> Self {
+        let name = expr_to_root_column_name(&self).unwrap();
+        let output_field = Some(Field::new(&name, DataType::Boolean));
         map_binary(
             self,
             other,
-            |left, other| {
+            move |left, other| {
                 left.is_in(&other).map(|ca| {
                     let mut s = ca.into_series();
-                    s.rename(left.name());
+                    s.rename(&name);
                     s
                 })
             },
-            Some(Field::new("", DataType::Boolean)),
+            output_field,
         )
     }
 
-    /// Get the year of a Date32/Date64
+    /// Get the year of a Date/Datetime
     #[cfg(feature = "temporal")]
     pub fn year(self) -> Expr {
         let function = move |s: Series| s.year().map(|ca| ca.into_series());
         self.map(function, GetOutput::from_type(DataType::UInt32))
     }
 
-    /// Get the month of a Date32/Date64
+    /// Get the month of a Date/Datetime
     #[cfg(feature = "temporal")]
     pub fn month(self) -> Expr {
         let function = move |s: Series| s.month().map(|ca| ca.into_series());
         self.map(function, GetOutput::from_type(DataType::UInt32))
     }
     /// Extract the week from the underlying Date representation.
-    /// Can be performed on Date32 and Date64
+    /// Can be performed on Date and Datetime
 
     /// Returns the ISO week number starting from 1.
     /// The return value ranges from 1 to 53. (The last week of year differs by years.)
@@ -1233,7 +1269,7 @@ impl Expr {
     }
 
     /// Extract the week day from the underlying Date representation.
-    /// Can be performed on Date32 and Date64.
+    /// Can be performed on Date and Datetime.
 
     /// Returns the weekday number where monday = 0 and sunday = 6
     #[cfg(feature = "temporal")]
@@ -1242,32 +1278,32 @@ impl Expr {
         self.map(function, GetOutput::from_type(DataType::UInt32))
     }
 
-    /// Get the month of a Date32/Date64
+    /// Get the month of a Date/Datetime
     #[cfg(feature = "temporal")]
     pub fn day(self) -> Expr {
         let function = move |s: Series| s.day().map(|ca| ca.into_series());
         self.map(function, GetOutput::from_type(DataType::UInt32))
     }
-    /// Get the ordinal_day of a Date32/Date64
+    /// Get the ordinal_day of a Date/Datetime
     #[cfg(feature = "temporal")]
     pub fn ordinal_day(self) -> Expr {
         let function = move |s: Series| s.ordinal_day().map(|ca| ca.into_series());
         self.map(function, GetOutput::from_type(DataType::UInt32))
     }
-    /// Get the hour of a Date64/Time64
+    /// Get the hour of a Datetime/Time64
     #[cfg(feature = "temporal")]
     pub fn hour(self) -> Expr {
         let function = move |s: Series| s.hour().map(|ca| ca.into_series());
         self.map(function, GetOutput::from_type(DataType::UInt32))
     }
-    /// Get the minute of a Date64/Time64
+    /// Get the minute of a Datetime/Time64
     #[cfg(feature = "temporal")]
     pub fn minute(self) -> Expr {
         let function = move |s: Series| s.minute().map(|ca| ca.into_series());
         self.map(function, GetOutput::from_type(DataType::UInt32))
     }
 
-    /// Get the second of a Date64/Time64
+    /// Get the second of a Datetime/Time64
     #[cfg(feature = "temporal")]
     pub fn second(self) -> Expr {
         let function = move |s: Series| s.second().map(|ca| ca.into_series());
@@ -1294,13 +1330,13 @@ impl Expr {
     #[cfg_attr(docsrs, doc(cfg(feature = "repeat_by")))]
     pub fn repeat_by(self, by: Expr) -> Expr {
         let function = |s: Series, by: Series| {
-            let by = by.cast_with_dtype(&DataType::UInt32)?;
+            let by = by.cast(&DataType::UInt32)?;
             Ok(s.repeat_by(by.u32()?).into_series())
         };
         map_binary_lazy_field(self, by, function, |_schema, _ctxt, l, _r| {
             Some(Field::new(
                 l.name(),
-                DataType::List(l.data_type().to_arrow()),
+                DataType::List(l.data_type().clone().into()),
             ))
         })
     }
@@ -1343,7 +1379,7 @@ impl Expr {
     /// use polars_lazy::prelude::*;
     ///
     /// fn example(df: LazyFrame) -> LazyFrame {
-    ///     df.select(vec![
+    ///     df.select([
     /// // even thought the alias yields a different column name,
     /// // `keep_name` will make sure that the original column name is used
     ///         col("*").alias("foo").keep_name()
@@ -1372,7 +1408,9 @@ impl Expr {
         }
     }
 
-    /// Exclude a column from a wildcard selection
+    /// Exclude a column from a wildcard/regex selection.
+    ///
+    /// You may also use regexes in the exclude as long as they start with `^` and end with `$`/
     ///
     /// # Example
     ///
@@ -1411,16 +1449,9 @@ impl Expr {
     /// [ChunkedArray::rolling_min](polars::prelude::ChunkWindow::rolling_min).
     #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
     #[cfg(feature = "rolling_window")]
-    pub fn rolling_min(
-        self,
-        window_size: u32,
-        weight: Option<&[f64]>,
-        ignore_null: bool,
-        min_periods: u32,
-    ) -> Expr {
-        let weight = weight.map(|v| v.to_vec());
+    pub fn rolling_min(self, options: RollingOptions) -> Expr {
         self.apply(
-            move |s| s.rolling_min(window_size, weight.as_deref(), ignore_null, min_periods),
+            move |s| s.rolling_min(options.clone()),
             GetOutput::same_type(),
         )
     }
@@ -1429,16 +1460,9 @@ impl Expr {
     /// [ChunkedArray::rolling_max](polars::prelude::ChunkWindow::rolling_max).
     #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
     #[cfg(feature = "rolling_window")]
-    pub fn rolling_max(
-        self,
-        window_size: u32,
-        weight: Option<&[f64]>,
-        ignore_null: bool,
-        min_periods: u32,
-    ) -> Expr {
-        let weight = weight.map(|v| v.to_vec());
+    pub fn rolling_max(self, options: RollingOptions) -> Expr {
         self.apply(
-            move |s| s.rolling_max(window_size, weight.as_deref(), ignore_null, min_periods),
+            move |s| s.rolling_max(options.clone()),
             GetOutput::same_type(),
         )
     }
@@ -1447,16 +1471,9 @@ impl Expr {
     /// [ChunkedArray::rolling_mean](polars::prelude::ChunkWindow::rolling_mean).
     #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
     #[cfg(feature = "rolling_window")]
-    pub fn rolling_mean(
-        self,
-        window_size: u32,
-        weight: Option<&[f64]>,
-        ignore_null: bool,
-        min_periods: u32,
-    ) -> Expr {
-        let weight = weight.map(|v| v.to_vec());
+    pub fn rolling_mean(self, options: RollingOptions) -> Expr {
         self.apply(
-            move |s| s.rolling_mean(window_size, weight.as_deref(), ignore_null, min_periods),
+            move |s| s.rolling_mean(options.clone()),
             GetOutput::same_type(),
         )
     }
@@ -1465,20 +1482,57 @@ impl Expr {
     /// [ChunkedArray::rolling_sum](polars::prelude::ChunkWindow::rolling_sum).
     #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
     #[cfg(feature = "rolling_window")]
-    pub fn rolling_sum(
-        self,
-        window_size: u32,
-        weight: Option<&[f64]>,
-        ignore_null: bool,
-        min_periods: u32,
-    ) -> Expr {
-        let weight = weight.map(|v| v.to_vec());
+    pub fn rolling_sum(self, options: RollingOptions) -> Expr {
         self.apply(
-            move |s| s.rolling_sum(window_size, weight.as_deref(), ignore_null, min_periods),
+            move |s| s.rolling_sum(options.clone()),
             GetOutput::same_type(),
         )
     }
 
+    /// Apply a rolling variance
+    #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
+    #[cfg(feature = "rolling_window")]
+    pub fn rolling_var(self, options: RollingOptions) -> Expr {
+        self.apply(
+            move |s| match s.dtype() {
+                DataType::Float32 => s.f32().unwrap().rolling_var(options.clone()),
+                DataType::Float64 => s.f64().unwrap().rolling_var(options.clone()),
+                _ => s
+                    .cast(&DataType::Float64)?
+                    .f64()
+                    .unwrap()
+                    .rolling_var(options.clone()),
+            },
+            GetOutput::map_field(|field| match field.data_type() {
+                DataType::Float64 => field.clone(),
+                DataType::Float32 => Field::new(field.name(), DataType::Float32),
+                _ => Field::new(field.name(), DataType::Float64),
+            }),
+        )
+    }
+
+    /// Apply a rolling std-dev
+    #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
+    #[cfg(feature = "rolling_window")]
+    pub fn rolling_std(self, options: RollingOptions) -> Expr {
+        self.apply(
+            move |s| match s.dtype() {
+                DataType::Float32 => s.f32().unwrap().rolling_std(options.clone()),
+                DataType::Float64 => s.f64().unwrap().rolling_std(options.clone()),
+                _ => s
+                    .cast(&DataType::Float64)?
+                    .f64()
+                    .unwrap()
+                    .rolling_std(options.clone()),
+            },
+            GetOutput::map_field(|field| match field.data_type() {
+                DataType::Float64 => field.clone(),
+                DataType::Float32 => Field::new(field.name(), DataType::Float32),
+                _ => Field::new(field.name(), DataType::Float64),
+            }),
+        )
+    }
+
     #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
     #[cfg(feature = "rolling_window")]
     /// Apply a custom function over a rolling/ moving window of the array.
@@ -1487,10 +1541,49 @@ impl Expr {
         self,
         window_size: usize,
         f: Arc<dyn Fn(&Series) -> Series + Send + Sync>,
+        output_type: GetOutput,
     ) -> Expr {
         self.apply(
             move |s| s.rolling_apply(window_size, f.as_ref()),
-            GetOutput::same_type(),
+            output_type,
+        )
+    }
+
+    #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
+    #[cfg(feature = "rolling_window")]
+    /// Apply a custom function over a rolling/ moving window of the array.
+    /// Prefer this over rolling_apply in case of floating point numbers as this is faster.
+    /// This has quite some dynamic dispatch, so prefer rolling_min, max, mean, sum over this.
+    pub fn rolling_apply_float<F>(self, window_size: usize, f: F) -> Expr
+    where
+        F: 'static + Fn(&Float64Chunked) -> Option<f64> + Send + Sync + Copy,
+    {
+        self.apply(
+            move |s| {
+                let out = match s.dtype() {
+                    DataType::Float64 => s
+                        .f64()
+                        .unwrap()
+                        .rolling_apply_float(window_size, f)
+                        .map(|ca| ca.into_series()),
+                    _ => s
+                        .cast(&DataType::Float64)?
+                        .f64()
+                        .unwrap()
+                        .rolling_apply_float(window_size, f)
+                        .map(|ca| ca.into_series()),
+                }?;
+                if let DataType::Float32 = s.dtype() {
+                    out.cast(&DataType::Float32)
+                } else {
+                    Ok(out)
+                }
+            },
+            GetOutput::map_field(|field| match field.data_type() {
+                DataType::Float64 => field.clone(),
+                DataType::Float32 => Field::new(field.name(), DataType::Float32),
+                _ => Field::new(field.name(), DataType::Float64),
+            }),
         )
     }
 
@@ -1535,6 +1628,18 @@ impl Expr {
             GetOutput::from_type(DataType::Float64),
         )
     }
+    #[cfg(feature = "concat_str")]
+    /// Concat the values into a string array.
+    /// # Arguments
+    ///
+    /// * `delimiter` - A string that will act as delimiter between values.
+    pub fn str_concat(self, delimiter: &str) -> Expr {
+        let delimiter = delimiter.to_owned();
+        self.apply(
+            move |s| Ok(s.str_concat(&delimiter).into_series()),
+            GetOutput::from_type(DataType::Utf8),
+        )
+    }
 }
 
 /// Create a Column Expression based on a column name.
@@ -1572,6 +1677,17 @@ pub fn cols(names: Vec<String>) -> Expr {
     Expr::Columns(names)
 }
 
+/// Select multiple columns by dtype.
+pub fn dtype_col(dtype: &DataType) -> Expr {
+    Expr::DtypeColumn(vec![dtype.clone()])
+}
+
+/// Select multiple columns by dtype.
+pub fn dtype_cols<DT: AsRef<[DataType]>>(dtype: DT) -> Expr {
+    let dtypes = dtype.as_ref().to_vec();
+    Expr::DtypeColumn(dtypes)
+}
+
 /// Count the number of values in this Expression.
 pub fn count(name: &str) -> Expr {
     match name {
@@ -1662,7 +1778,7 @@ where
             let mut acc = series.pop().unwrap();
 
             for s in series {
-                acc = f(acc, s)?
+                acc = f(acc, s)?;
             }
             Ok(acc)
         }) as Arc<dyn SeriesUdf>);
@@ -1772,7 +1888,6 @@ make_literal!(u8, UInt8);
 #[cfg(feature = "dtype-u16")]
 make_literal!(u16, UInt16);
 make_literal!(u32, UInt32);
-#[cfg(feature = "dtype-u64")]
 make_literal!(u64, UInt64);
 
 /// The literal Null
@@ -1784,14 +1899,14 @@ impl Literal for Null {
     }
 }
 
-#[cfg(all(feature = "temporal", feature = "dtype-date64"))]
+#[cfg(all(feature = "temporal", feature = "dtype-datetime"))]
 impl Literal for NaiveDateTime {
     fn lit(self) -> Expr {
         Expr::Literal(LiteralValue::DateTime(self))
     }
 }
 
-#[cfg(all(feature = "temporal", feature = "dtype-date64"))]
+#[cfg(all(feature = "temporal", feature = "dtype-datetime"))]
 impl Literal for NaiveDate {
     fn lit(self) -> Expr {
         Expr::Literal(LiteralValue::DateTime(self.and_hms(0, 0, 0)))
@@ -1901,6 +2016,85 @@ impl Rem for Expr {
     }
 }
 
+/// Apply a function/closure over multiple columns once the logical plan get executed.
+///
+/// This function is very similar to `[apply_mul]`, but differs in how it handles aggregations.
+///
+///  * `map_mul` should be used for operations that are independent of groups, e.g. `multiply * 2`, or `raise to the power`
+///  * `apply_mul` should be used for operations that work on a group of data. e.g. `sum`, `count`, etc.
+///
+/// It is the responsibility of the caller that the schema is correct by giving
+/// the correct output_type. If None given the output type of the input expr is used.
+pub fn map_mul<F, E>(function: F, expr: E, output_type: GetOutput) -> Expr
+where
+    F: Fn(&mut [Series]) -> Result<Series> + 'static + Send + Sync,
+    E: AsRef<[Expr]>,
+{
+    let input = expr.as_ref().to_vec();
+
+    Expr::Function {
+        input,
+        function: NoEq::new(Arc::new(function)),
+        output_type,
+        options: FunctionOptions {
+            collect_groups: ApplyOptions::ApplyFlat,
+            input_wildcard_expansion: false,
+        },
+    }
+}
+
+/// Apply a function/closure over multiple columns once the logical plan get executed.
+///
+/// This function is very similar to `[apply_mul]`, but differs in how it handles aggregations.
+///
+///  * `map_mul` should be used for operations that are independent of groups, e.g. `multiply * 2`, or `raise to the power`
+///  * `apply_mul` should be used for operations that work on a group of data. e.g. `sum`, `count`, etc.
+///  * `map_list_mul` should be used when the function expects a list aggregated series.
+pub fn map_list_mul<F, E>(function: F, expr: E, output_type: GetOutput) -> Expr
+where
+    F: Fn(&mut [Series]) -> Result<Series> + 'static + Send + Sync,
+    E: AsRef<[Expr]>,
+{
+    let input = expr.as_ref().to_vec();
+
+    Expr::Function {
+        input,
+        function: NoEq::new(Arc::new(function)),
+        output_type,
+        options: FunctionOptions {
+            collect_groups: ApplyOptions::ApplyList,
+            input_wildcard_expansion: false,
+        },
+    }
+}
+
+/// Apply a function/closure over the groups of multiple columns. This should only be used in a groupby aggregation.
+///
+/// It is the responsibility of the caller that the schema is correct by giving
+/// the correct output_type. If None given the output type of the input expr is used.
+///
+/// This difference with `[map_mul]` is that `[apply_mul]` will create a separate `[Series]` per group.
+///
+/// * `[map_mul]` should be used for operations that are independent of groups, e.g. `multiply * 2`, or `raise to the power`
+/// * `[apply_mul]` should be used for operations that work on a group of data. e.g. `sum`, `count`, etc.
+pub fn apply_mul<F, E>(function: F, expr: E, output_type: GetOutput) -> Expr
+where
+    F: Fn(&mut [Series]) -> Result<Series> + 'static + Send + Sync,
+    E: AsRef<[Expr]>,
+{
+    let input = expr.as_ref().to_vec();
+
+    Expr::Function {
+        input,
+        function: NoEq::new(Arc::new(function)),
+        output_type,
+        options: FunctionOptions {
+            collect_groups: ApplyOptions::ApplyGroups,
+            input_wildcard_expansion: false,
+        },
+    }
+}
+
 #[cfg(test)]
 mod test {
     use super::*;
