@@ -3,6 +3,7 @@
 use polars_core::datatypes::PlHashMap;
 use polars_core::frame::hash_join::JoinType;
 use polars_core::prelude::*;
+#[cfg(feature = "dtype-categorical")]
 use polars_core::toggle_string_cache;
 use std::sync::Arc;
 
@@ -38,6 +39,7 @@ pub struct LazyCsvReader<'a> {
     schema_overwrite: Option<&'a Schema>,
     low_memory: bool,
     comment_char: Option<u8>,
+    quote_char: Option<u8>,
     null_values: Option<NullValues>,
 }
 
@@ -56,6 +58,7 @@ impl<'a> LazyCsvReader<'a> {
             schema_overwrite: None,
             low_memory: false,
             comment_char: None,
+            quote_char: Some(b'"'),
             null_values: None,
         }
     }
@@ -110,6 +113,12 @@ impl<'a> LazyCsvReader<'a> {
         self
     }
 
+    /// Set the `char` used as quote char. The default is `b'"'`. If set to `[None]` quoting is disabled.
+    pub fn with_quote_char(mut self, quote: Option<u8>) -> Self {
+        self.quote_char = quote;
+        self
+    }
+
     /// Set values that will be interpreted as missing/ null.
     pub fn with_null_values(mut self, null_values: Option<NullValues>) -> Self {
         self.null_values = null_values;
@@ -141,6 +150,7 @@ impl<'a> LazyCsvReader<'a> {
             self.schema_overwrite,
             self.low_memory,
             self.comment_char,
+            self.quote_char,
             self.null_values,
         )
         .build()
@@ -150,10 +160,14 @@ impl<'a> LazyCsvReader<'a> {
     }
 }
 
-#[derive(Copy, Clone, Debug)]
+#[derive(Clone, Debug)]
 pub struct JoinOptions {
     pub allow_parallel: bool,
     pub force_parallel: bool,
+    pub how: JoinType,
+    pub suffix: Option<String>,
+    pub asof_by_left: Vec<String>,
+    pub asof_by_right: Vec<String>,
 }
 
 impl Default for JoinOptions {
@@ -161,6 +175,10 @@ impl Default for JoinOptions {
         JoinOptions {
             allow_parallel: true,
             force_parallel: false,
+            how: JoinType::Left,
+            suffix: None,
+            asof_by_left: vec![],
+            asof_by_right: vec![],
         }
     }
 }
@@ -179,21 +197,12 @@ impl IntoLazy for DataFrame {
 /// Lazy abstraction over an eager `DataFrame`.
 /// It really is an abstraction over a logical plan. The methods of this struct will incrementally
 /// modify a logical plan until output is requested (via [collect](crate::frame::LazyFrame::collect))
-#[derive(Clone)]
+#[derive(Clone, Default)]
 pub struct LazyFrame {
     pub(crate) logical_plan: LogicalPlan,
     pub(crate) opt_state: OptState,
 }
 
-impl Default for LazyFrame {
-    fn default() -> Self {
-        LazyFrame {
-            logical_plan: LogicalPlan::default(),
-            opt_state: Default::default(),
-        }
-    }
-}
-
 impl From<LogicalPlan> for LazyFrame {
     fn from(plan: LogicalPlan) -> Self {
         Self {
@@ -434,6 +443,49 @@ impl LazyFrame {
         init.map(f, Some(AllowedOptimizations::default()), Some(schema))
     }
 
+    /// Rename columns in the DataFrame. This does not preserve ordering.
+    pub fn rename<I, J, T, S>(self, existing: I, new: J) -> Self
+    where
+        I: IntoIterator<Item = T> + Clone,
+        J: IntoIterator<Item = S>,
+        T: AsRef<str>,
+        S: AsRef<str>,
+    {
+        let existing: Vec<String> = existing
+            .into_iter()
+            .map(|name| name.as_ref().to_string())
+            .collect();
+
+        self.with_columns(
+            existing
+                .iter()
+                .zip(new)
+                .map(|(old, new)| col(old).alias(new.as_ref()))
+                .collect(),
+        )
+        .drop_columns_impl(&existing)
+    }
+
+    /// Removes columns from the DataFrame.
+    /// Note that its better to only select the columns you need
+    /// and let the projection pushdown optimize away the unneeded columns.
+    pub fn drop_columns<I, T>(self, columns: I) -> Self
+    where
+        I: IntoIterator<Item = T>,
+        T: AsRef<str>,
+    {
+        let columns: Vec<String> = columns
+            .into_iter()
+            .map(|name| name.as_ref().to_string())
+            .collect();
+        self.drop_columns_impl(&columns)
+    }
+
+    #[allow(clippy::ptr_arg)]
+    fn drop_columns_impl(self, columns: &Vec<String>) -> Self {
+        self.select_local(vec![col("*").exclude(columns)])
+    }
+
     /// Shift the values by a given period and fill the parts that will be empty due to this operation
     /// with `Nones`.
     ///
@@ -592,20 +644,21 @@ impl LazyFrame {
     /// use polars_lazy::prelude::*;
     ///
     /// fn example(df: DataFrame) -> Result<DataFrame> {
-    ///       df.lazy()
-    ///         .groupby(vec![col("foo")])
-    ///         .agg(vec!(col("bar").sum(),
-    ///                   col("ham").mean().alias("avg_ham")))
-    ///         .collect()
+    ///     df.lazy()
+    ///       .groupby([col("foo")])
+    ///       .agg([col("bar").sum(), col("ham").mean().alias("avg_ham")])
+    ///       .collect()
     /// }
     /// ```
     pub fn collect(self) -> Result<DataFrame> {
+        #[cfg(feature = "dtype-categorical")]
         let use_string_cache = self.opt_state.global_string_cache;
         let mut expr_arena = Arena::with_capacity(256);
         let mut lp_arena = Arena::with_capacity(128);
         let lp_top = self.optimize(&mut lp_arena, &mut expr_arena)?;
 
         // if string cache was already set, we skip this and global settings are respected
+        #[cfg(feature = "dtype-categorical")]
         if use_string_cache {
             toggle_string_cache(use_string_cache);
         }
@@ -615,6 +668,7 @@ impl LazyFrame {
 
         let state = ExecutionState::new();
         let out = physical_plan.execute(&state);
+        #[cfg(feature = "dtype-categorical")]
         if use_string_cache {
             toggle_string_cache(!use_string_cache);
         }
@@ -693,8 +747,8 @@ impl LazyFrame {
     ///
     /// fn example(df: DataFrame) -> LazyFrame {
     ///       df.lazy()
-    ///        .groupby(vec![col("date")])
-    ///        .agg(vec![
+    ///        .groupby([col("date")])
+    ///        .agg([
     ///            col("rain").min(),
     ///            col("rain").sum(),
     ///            col("rain").quantile(0.5).alias("median_rain"),
@@ -702,23 +756,23 @@ impl LazyFrame {
     ///        .sort("date", false)
     /// }
     /// ```
-    pub fn groupby(self, by: Vec<Expr>) -> LazyGroupBy {
+    pub fn groupby<E: AsRef<[Expr]>>(self, by: E) -> LazyGroupBy {
         let opt_state = self.get_opt_state();
         LazyGroupBy {
             logical_plan: self.logical_plan,
             opt_state,
-            keys: by,
+            keys: by.as_ref().to_vec(),
             maintain_order: false,
         }
     }
 
     /// Similar to groupby, but order of the DataFrame is maintained.
-    pub fn stable_groupby(self, by: Vec<Expr>) -> LazyGroupBy {
+    pub fn stable_groupby<E: AsRef<[Expr]>>(self, by: E) -> LazyGroupBy {
         let opt_state = self.get_opt_state();
         LazyGroupBy {
             logical_plan: self.logical_plan,
             opt_state,
-            keys: by,
+            keys: by.as_ref().to_vec(),
             maintain_order: true,
         }
     }
@@ -805,6 +859,7 @@ impl LazyFrame {
             .finish()
     }
 
+    /// Control more join options with the join builder.
     pub fn join_builder(self) -> JoinBuilder {
         JoinBuilder::new(self)
     }
@@ -989,6 +1044,23 @@ impl LazyFrame {
             .build();
         Self::from_logical_plan(lp, opt_state)
     }
+
+    /// Add a new column at index 0 that counts the rows.
+    pub fn with_row_count(self, name: &str) -> LazyFrame {
+        let schema = self.schema();
+
+        let mut fields = schema.fields().clone();
+        fields.insert(0, Field::new(name, DataType::UInt32));
+        let new_schema = Schema::new(fields);
+
+        let name = name.to_owned();
+
+        self.map(
+            move |df: DataFrame| df.with_row_count(&name),
+            Some(AllowedOptimizations::default()),
+            Some(new_schema),
+        )
+    }
 }
 
 /// Utility struct for lazy groupby operation.
@@ -1013,8 +1085,8 @@ impl LazyGroupBy {
     ///
     /// fn example(df: DataFrame) -> LazyFrame {
     ///       df.lazy()
-    ///        .groupby(vec![col("date")])
-    ///        .agg(vec![
+    ///        .groupby([col("date")])
+    ///        .agg([
     ///            col("rain").min(),
     ///            col("rain").sum(),
     ///            col("rain").quantile(0.5).alias("median_rain"),
@@ -1022,7 +1094,7 @@ impl LazyGroupBy {
     ///        .sort("date", false)
     /// }
     /// ```
-    pub fn agg(self, aggs: Vec<Expr>) -> LazyFrame {
+    pub fn agg<E: AsRef<[Expr]>>(self, aggs: E) -> LazyFrame {
         let lp = LogicalPlanBuilder::from(self.logical_plan)
             .groupby(Arc::new(self.keys), aggs, None, self.maintain_order)
             .build();
@@ -1038,7 +1110,7 @@ impl LazyGroupBy {
             .flatten()
             .collect::<Vec<_>>();
 
-        self.agg(vec![col("*").exclude(&keys).head(n).list().keep_name()])
+        self.agg([col("*").exclude(&keys).head(n).list().keep_name()])
             .explode(vec![col("*").exclude(&keys)])
     }
 
@@ -1051,7 +1123,7 @@ impl LazyGroupBy {
             .flatten()
             .collect::<Vec<_>>();
 
-        self.agg(vec![col("*").exclude(&keys).tail(n).list().keep_name()])
+        self.agg([col("*").exclude(&keys).tail(n).list().keep_name()])
             .explode(vec![col("*").exclude(&keys)])
     }
 
@@ -1081,6 +1153,9 @@ pub struct JoinBuilder {
     right_on: Vec<Expr>,
     allow_parallel: bool,
     force_parallel: bool,
+    suffix: Option<String>,
+    asof_by_left: Vec<String>,
+    asof_by_right: Vec<String>,
 }
 impl JoinBuilder {
     fn new(lf: LazyFrame) -> Self {
@@ -1092,36 +1167,62 @@ impl JoinBuilder {
             right_on: vec![],
             allow_parallel: true,
             force_parallel: false,
+            suffix: None,
+            asof_by_left: vec![],
+            asof_by_right: vec![],
         }
     }
 
+    /// The table to join with.
     pub fn with(mut self, other: LazyFrame) -> Self {
         self.other = Some(other);
         self
     }
 
+    /// Select the join type.
     pub fn how(mut self, how: JoinType) -> Self {
         self.how = how;
         self
     }
 
+    /// The columns you want to join the left table on.
     pub fn left_on(mut self, on: Vec<Expr>) -> Self {
         self.left_on = on;
         self
     }
 
+    /// The columns you want to join the right table on.
     pub fn right_on(mut self, on: Vec<Expr>) -> Self {
         self.right_on = on;
         self
     }
+    /// Allow parallel table evaluation.
     pub fn allow_parallel(mut self, allow: bool) -> Self {
         self.allow_parallel = allow;
         self
     }
+
+    /// Force parallel table evaluation.
     pub fn force_parallel(mut self, allow: bool) -> Self {
         self.allow_parallel = allow;
         self
     }
+
+    /// Suffix to add duplicate column names in join.
+    /// Defaults to `"_right"`.
+    pub fn suffix(mut self, suffix: String) -> Self {
+        self.suffix = Some(suffix);
+        self
+    }
+
+    /// Set the `by` subgrouper of an asof join.
+    pub fn asof_by(mut self, left_by: Vec<String>, right_by: Vec<String>) -> Self {
+        self.asof_by_left = left_by;
+        self.asof_by_right = right_by;
+        self
+    }
+
+    /// Finish builder
     pub fn finish(self) -> LazyFrame {
         let opt_state = self.lf.opt_state;
 
@@ -1130,11 +1231,16 @@ impl JoinBuilder {
             .get_plan_builder()
             .join(
                 self.other.expect("with not set").logical_plan,
-                self.how,
                 self.left_on,
                 self.right_on,
-                self.allow_parallel,
-                self.force_parallel,
+                JoinOptions {
+                    allow_parallel: self.allow_parallel,
+                    force_parallel: self.force_parallel,
+                    how: self.how,
+                    suffix: self.suffix,
+                    asof_by_left: self.asof_by_left,
+                    asof_by_right: self.asof_by_right,
+                },
             )
             .build();
         LazyFrame::from_logical_plan(lp, opt_state)
