@@ -1,6 +1,6 @@
 use crate::logical_plan::Context;
 use crate::prelude::*;
-use crate::utils::{aexpr_to_root_names, aexpr_to_root_nodes, check_down_node, has_aexpr};
+use crate::utils::{aexpr_to_root_names, aexpr_to_root_nodes, check_input_node, has_aexpr};
 use polars_core::{datatypes::PlHashSet, prelude::*};
 
 fn init_vec() -> Vec<Node> {
@@ -46,7 +46,7 @@ fn split_acc_projections(
     } else {
         let (acc_projections, local_projections): (Vec<Node>, Vec<Node>) = acc_projections
             .into_iter()
-            .partition(|expr| check_down_node(*expr, down_schema, expr_arena));
+            .partition(|expr| check_input_node(*expr, down_schema, expr_arena));
         let mut names = init_set();
         for proj in &acc_projections {
             for name in aexpr_to_root_names(*proj, expr_arena) {
@@ -118,13 +118,13 @@ impl ProjectionPushDown {
         let root_projections = aexpr_to_root_nodes(proj, expr_arena);
 
         for (name, root_projection) in names.into_iter().zip(root_projections) {
-            if check_down_node(root_projection, schema_left, expr_arena)
+            if check_input_node(root_projection, schema_left, expr_arena)
                 && names_left.insert(name.clone())
             {
                 pushdown_left.push(proj);
                 pushed_at_least_one = true;
             }
-            if check_down_node(root_projection, schema_right, expr_arena)
+            if check_input_node(root_projection, schema_right, expr_arena)
                 && names_right.insert(name)
             {
                 pushdown_right.push(proj);
@@ -270,7 +270,7 @@ impl ProjectionPushDown {
                 // projection from a wildcard may be dropped if the schema changes due to the optimization
                 let proj = expr
                     .into_iter()
-                    .filter(|e| check_down_node(*e, schema, expr_arena))
+                    .filter(|e| check_input_node(*e, schema, expr_arena))
                     .collect();
                 Ok(ALogicalPlanBuilder::new(input, expr_arena, lp_arena)
                     .project_local(proj)
@@ -493,6 +493,7 @@ impl ProjectionPushDown {
                 aggs,
                 apply,
                 schema,
+                maintain_order,
             } => {
                 // the custom function may need all columns so we do the projections here.
                 if let Some(f) = apply {
@@ -502,6 +503,7 @@ impl ProjectionPushDown {
                         aggs,
                         schema,
                         apply: Some(f),
+                        maintain_order,
                     };
                     let input = lp_arena.add(lp);
 
@@ -535,8 +537,12 @@ impl ProjectionPushDown {
                         expr_arena,
                     )?;
 
-                    let builder = ALogicalPlanBuilder::new(input, expr_arena, lp_arena)
-                        .groupby(keys, aggs, apply);
+                    let builder = ALogicalPlanBuilder::new(input, expr_arena, lp_arena).groupby(
+                        keys,
+                        aggs,
+                        apply,
+                        maintain_order,
+                    );
                     Ok(builder.build())
                 }
             }
@@ -545,9 +551,7 @@ impl ProjectionPushDown {
                 input_right,
                 left_on,
                 right_on,
-                how,
-                allow_par,
-                force_par,
+                options,
                 ..
             } => {
                 let mut pushdown_left = init_vec();
@@ -556,9 +560,14 @@ impl ProjectionPushDown {
                 let mut names_right = init_set();
                 let mut local_projection = init_vec();
 
-                // if there are no projections we don't have to do anything
+                // if there are no projections we don't have to do anything (all columns are projected)
                 // otherwise we build local projections to sort out proper column names due to the
                 // join operation
+                //
+                // Joins on columns with different names, for example
+                // left_on = "a", right_on = "b
+                // will remove the name "b" (it is "a" now). That columns should therefore not
+                // be added to a local projection.
                 if !acc_projections.is_empty() {
                     let schema_left = lp_arena.get(input_left).schema(lp_arena);
                     let schema_right = lp_arena.get(input_right).schema(lp_arena);
@@ -582,9 +591,7 @@ impl ProjectionPushDown {
                             &mut names_right,
                             expr_arena,
                         );
-                        if !local_projection.contains(e) {
-                            local_projection.push(*e)
-                        }
+                        // we don't add right column names to local_projection as they are removed
                     }
 
                     for proj in acc_projections {
@@ -620,12 +627,13 @@ impl ProjectionPushDown {
                             let root_column_name =
                                 aexpr_to_root_names(proj, expr_arena).pop().unwrap();
 
+                            let suffix = options.suffix.as_deref().unwrap_or("_right");
                             // If _right suffix exists we need to push a projection down without this
                             // suffix.
-                            if root_column_name.ends_with("_right") {
+                            if root_column_name.ends_with(suffix) {
                                 // downwards name is the name without the _right i.e. "foo".
                                 let (downwards_name, _) = root_column_name
-                                    .split_at(root_column_name.len() - "_right".len());
+                                    .split_at(root_column_name.len() - suffix.len());
 
                                 let downwards_name_column =
                                     expr_arena.add(AExpr::Column(Arc::new(downwards_name.into())));
@@ -637,7 +645,7 @@ impl ProjectionPushDown {
                                 // locally we project and alias
                                 let projection = expr_arena.add(AExpr::Alias(
                                     downwards_name_column,
-                                    Arc::new(format!("{}_right", downwards_name)),
+                                    Arc::new(format!("{}{}", downwards_name, suffix)),
                                 ));
                                 local_projection.push(projection);
                             }
@@ -675,11 +683,9 @@ impl ProjectionPushDown {
 
                 let builder = ALogicalPlanBuilder::new(input_left, expr_arena, lp_arena).join(
                     input_right,
-                    how,
                     left_on,
                     right_on,
-                    allow_par,
-                    force_par,
+                    options,
                 );
                 Ok(self.finish_node(local_projection, builder))
             }
