@@ -1,11 +1,9 @@
 use crate::logical_plan::Context;
-use crate::physical_plan::expressions::binary::binary_check_group_tuples;
 use crate::physical_plan::state::ExecutionState;
 use crate::physical_plan::PhysicalAggregation;
 use crate::prelude::*;
 use polars_core::frame::groupby::GroupTuples;
 use polars_core::{prelude::*, POOL};
-use std::borrow::Cow;
 use std::sync::Arc;
 
 pub(crate) struct BinaryFunctionExpr {
@@ -13,9 +11,14 @@ pub(crate) struct BinaryFunctionExpr {
     pub(crate) input_b: Arc<dyn PhysicalExpr>,
     pub(crate) function: NoEq<Arc<dyn SeriesBinaryUdf>>,
     pub(crate) output_field: NoEq<Arc<dyn BinaryUdfOutputField>>,
+    pub(crate) expr: Expr,
 }
 
 impl PhysicalExpr for BinaryFunctionExpr {
+    fn as_expression(&self) -> &Expr {
+        &self.expr
+    }
+
     fn evaluate(&self, df: &DataFrame, state: &ExecutionState) -> Result<Series> {
         let (series_a, series_b) = POOL.install(|| {
             rayon::join(
@@ -49,32 +52,57 @@ impl PhysicalExpr for BinaryFunctionExpr {
         df: &DataFrame,
         groups: &'a GroupTuples,
         state: &ExecutionState,
-    ) -> Result<(Series, Cow<'a, GroupTuples>)> {
+    ) -> Result<AggregationContext<'a>> {
         let (series_a, series_b) = POOL.install(|| {
             rayon::join(
                 || self.input_a.evaluate_on_groups(df, groups, state),
                 || self.input_b.evaluate_on_groups(df, groups, state),
             )
         });
-        let (series_a, groups_a) = series_a?;
-        let (series_b, groups_b) = series_b?;
+        let mut ac_l = series_a?;
+        let mut ac_r = series_b?;
 
-        let name = self
-            .output_field
-            .get_field(
-                &df.schema(),
-                Context::Default,
-                &Field::new(series_a.name(), series_a.dtype().clone()),
-                &Field::new(series_b.name(), series_b.dtype().clone()),
-            )
-            .map(|fld| fld.name().clone())
-            .unwrap_or_else(|| "binary_function".to_string());
+        if !ac_l.can_combine(&ac_r) {
+            return Err(PolarsError::InvalidOperation(
+                "\
+            cannot combine this binary function, the groups do not match"
+                    .into(),
+            ));
+        }
 
-        let out = self.function.call_udf(series_a, series_b).map(|mut s| {
-            s.rename(&name);
-            s
-        })?;
-        binary_check_group_tuples(out, groups_a, groups_b)
+        // keep track of the output lengths. If they are all unit length,
+        // we can explode the array as it would have the same length as the no. of groups
+        // if it is not all unit length it should remain a listarray
+        let mut all_unit_length = true;
+
+        let ca = ac_l
+            .aggregated()
+            .list()
+            .unwrap()
+            .into_iter()
+            .zip(ac_r.aggregated().list().unwrap())
+            .map(|(opt_a, opt_b)| match (opt_a, opt_b) {
+                (Some(a), Some(b)) => {
+                    let out = self.function.call_udf(a, b).ok();
+
+                    if let Some(s) = &out {
+                        if s.len() != 1 {
+                            all_unit_length = false;
+                        }
+                    }
+                    out
+                }
+                _ => None,
+            })
+            .collect::<ListChunked>();
+
+        ac_l.combine_groups(ac_r);
+        if all_unit_length {
+            ac_l.with_series(ca.explode()?);
+            return Ok(ac_l);
+        }
+        ac_l.with_series(ca.into_series());
+        Ok(ac_l)
     }
 
     fn to_field(&self, input_schema: &Schema) -> Result<Field> {
@@ -98,12 +126,12 @@ impl PhysicalAggregation for BinaryFunctionExpr {
         let (agg_a, agg_b): (Result<Series>, Result<Series>) = POOL.install(|| {
             rayon::join(
                 || {
-                    let a = self.input_a.evaluate(df, state)?;
-                    Ok(a.agg_list(groups).expect("no data"))
+                    let mut ac = self.input_a.evaluate_on_groups(df, groups, state)?;
+                    Ok(ac.aggregated().into_owned())
                 },
                 || {
-                    let b = self.input_b.evaluate(df, state)?;
-                    Ok(b.agg_list(groups).expect("no data"))
+                    let mut ac = self.input_b.evaluate_on_groups(df, groups, state)?;
+                    Ok(ac.aggregated().into_owned())
                 },
             )
         });
