@@ -1,5 +1,6 @@
 use crate::physical_plan::state::ExecutionState;
 use crate::prelude::*;
+use polars_core::frame::groupby::GroupTuples;
 use polars_core::prelude::*;
 use std::sync::Arc;
 
@@ -26,6 +27,21 @@ impl PhysicalExpr for IsNullExpr {
         let series = self.physical_expr.evaluate(df, state)?;
         Ok(series.is_null().into_series())
     }
+    #[allow(clippy::ptr_arg)]
+    fn evaluate_on_groups<'a>(
+        &self,
+        df: &DataFrame,
+        groups: &'a GroupTuples,
+        state: &ExecutionState,
+    ) -> Result<AggregationContext<'a>> {
+        let mut ac = self.physical_expr.evaluate_on_groups(df, groups, state)?;
+        let s = ac.flat();
+        let s = s.is_null().into_series();
+        ac.with_series(s);
+
+        Ok(ac)
+    }
+
     fn to_field(&self, _input_schema: &Schema) -> Result<Field> {
         Ok(Field::new("is_null", DataType::Boolean))
     }
